// Copyright (c) 1997 James Clark
// See the file COPYING for copying permission.


#include "config.h"
#include "SdNode.h"

#ifdef SP_NAMESPACE
namespace SP_NAMESPACE {
#endif

#ifdef GROVE_NAMESPACE
using namespace GROVE_NAMESPACE;
#endif

const Node::IID SdNode::iid = "SdNode";

#ifdef SP_NAMESPACE
}
#endif
